def c3d2zoo_data(c3d_obj):
    """
    Converts an ezc3d C3D object to zoo format.

    Returns:
    - data (dict): Zoo dictionary with 'line' and 'event' fields per channel.
    """
    data = {}

    if 'points' in c3d_obj['data']:
        points = c3d_obj['data']['points']  # shape: (4, n_markers, n_frames)
        labels = c3d_obj['parameters']['POINT']['LABELS']['value']
        for i, label in enumerate(labels):
            line_data = points[:3, i, :].T  # shape: (frames, 3)
            data[label] = {
                'line': line_data,
                'event': {}  # empty for now
            }

        params = c3d_obj['parameters']
        video_freq = c3d_obj['parameters']['POINT']['RATE']['value'][0]
        if 'EVENT' in params and 'TIMES' in params['EVENT']:
            times_array = params['EVENT']['TIMES']['value']
            frames = times_array[1]  # second row = frames (or time, depending on C3D file)

            # Extract sides, types, subjects
            contexts = params['EVENT']['CONTEXTS']['value'] if 'CONTEXTS' in params['EVENT'] else ['']
            labels = params['EVENT']['LABELS']['value']
            subjects = params['EVENT']['SUBJECTS']['value'] if 'SUBJECTS' in params['EVENT'] else ['']

            events = {}

            for i in range(len(labels)):
                side = contexts[i].strip()
                label = labels[i].strip()
                subject = subjects[i].strip()

                # Event channel name: e.g. 'Right_FootStrike' -> 'RightFootStrike'
                event_name = f"{side}_{label}".replace(' ', '')
                event_name = ''.join(c for c in event_name if c.isalnum() or c == '_')  # make it a valid field name

                if event_name not in events:
                    events[event_name] = []

                events[event_name].append(frames[i])  # This is in seconds or frame number?

            original_start = 1

            for event_name, time_list in events.items():
                # Clean and sort times
                valid_times = sorted([t for t in time_list if t != 0])
                for j, time_val in enumerate(valid_times):
                    frame = round(time_val * video_freq) - original_start + 1  # MATLAB logic
                    key_name = f"{event_name}{j + 1}"

                    # Place in correct channel
                    if 'SACR' in data:
                        data['SACR']['event'][key_name] = [frame, 0, 0]
                    else:
                        data[labels[0]]['event'][key_name] = [frame, 0, 0]

    # todo add relevant meta data to zoosystem
    data['zoosystem'] = params['EVENT']

    return data



import numpy as np
from scipy.signal import hilbert


def phase_angle_line(r):
    """
    Computes the phase angle for a single kinematic waveform using the Hilbert transform method.

    Parameters:
    r : array_like
        (n, 1) array of kinematic data (e.g., joint or segment angle)

    Returns:
    PA_data : ndarray
        1D array of phase angle (in degrees) computed from input using the Hilbert transform.

    Reference:
    Lamb and Stöckl (2014). "On the use of continuous relative phase..."
    Clinical Biomechanics. https://doi.org/10.1016/j.clinbiomech.2014.03.008
    """

    # Step 1: Center the data around zero as per Lamb and Stöckl eq. 11
    cdata = r - np.min(r) - (np.max(r) - np.min(r)) / 2

    # Step 2: Hilbert transform
    X = hilbert(cdata)

    # Step 3: Phase angle calculation
    PA = np.rad2deg(np.arctan2(np.imag(X), np.real(X)))

    return PA


if __name__ == '__main__':
    # -------TESTING--------
    import os
    from biomechzoo.utils.zload import zload
    from matplotlib import pyplot as plt
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC032A18_exploded.zoo')
    data = zload(fl)
    print(data)
    r = data['RKneeAngles_x']['line']
    phase_angle = phase_angle_line(r)
    plt.plot(phase_angle)
    plt.show()




import os
import argparse
import matplotlib.pyplot as plt
from biomechzoo.mvn.load_mvnx import load_mvnx


# Convert mvnx file to python data
def main(file_name):
    # Check for file existence
    if not os.path.isfile(file_name):
        raise Exception("File %s could not be found" % file_name)

    tokens = file_name.lower().split('.')
    extension = tokens[-1]

    # Check for file extension
    if not extension == 'mvnx':
        raise Exception("File must be an .mvnx file")

    # Load data
    mvnx_file = load_mvnx(file_name)

    # Read some basic data from the file
    comments = mvnx_file.comments
    frame_rate = mvnx_file.frame_rate
    configuration = mvnx_file.configuration
    original_file_name = mvnx_file.original_file_name
    recording_date = mvnx_file.recording_date
    actor_name = mvnx_file.actor_name
    frame_count = mvnx_file.frame_count
    version = mvnx_file.version
    segment_count = mvnx_file.segment_count
    joint_count = mvnx_file.joint_count

    # Read the data from the structure e.g. first segment
    idx = 0
    segment_name = mvnx_file.segment_name_from_index(idx)
    segment_pos = mvnx_file.get_segment_pos(idx)

    # Alternatively, use the generic method get_data() with the data set and field. E.g.:
    # segment_pos = mvnx_file.get_data('segment_data', 'pos', idx)

    if segment_pos:
        # Plot position of a segment
        plt.figure(0)
        plt.plot(segment_pos)
        plt.xlabel('frames')
        plt.ylabel('Position in the global frame')
        plt.title('Position of ' + segment_name + ' segment')
        plt.legend(['x', 'y', 'z'])
        plt.draw()

        # Plot 3D displacement of a segment
        x, y, z = map(list, zip(*[[frame[0], frame[1], frame[2]] for frame in segment_pos]))
        plt.figure(1)
        plt.axes(projection="3d")
        plt.plot(x, y, z)
        plt.xlabel('frames')
        plt.title('Position of ' + segment_name + ' segment in 3D')
        plt.draw()

        plt.show()


if __name__ == '__main__':

    # Program entry point
    parser = argparse.ArgumentParser()
    parser.add_argument('--mvnx_file', required=True, type=str, help='The MVNX file to load', nargs='?')
    args = parser.parse_args()

    try:
        main(args.mvnx_file)
    except Exception as e:
        print("Error: %s" % e)



from scipy.io import savemat
import inspect
import os


def zsave(fl, data, inplace=True, out_folder=None, root_folder=None):
    """
    Save zoo data to .zoo file (MAT format)

    Arguments:
        fl (str): Full path to original .zoo file
        data (dict): Zoo data to save
        inplace (bool): Whether to overwrite original file
        out_folder (str or None): If not inplace, output folder name (relative to root_folder or file location)
        root_folder (str or None): Optional base directory for saving when inplace=False
    """
    # Get caller function name for logging
    caller_name = inspect.stack()[1].function

    # Initialize zoosystem and processing history
    zoosystem = data.get('zoosystem', {})
    processing = zoosystem.get('Processing', [])
    if not isinstance(processing, list):
        processing = [processing]
    processing.append(caller_name)
    zoosystem['Processing'] = processing
    data['zoosystem'] = zoosystem

    # Determine save path
    if inplace:
        save_path = fl
    else:
        filename = os.path.basename(fl)
        if out_folder is None:
            out_folder = 'processed'

        if root_folder is None:
            root_folder = os.path.dirname(fl)

        root_path = os.path.dirname(root_folder)
        in_folder = os.path.basename(root_folder)
        out_dir = os.path.join(root_path, out_folder)
        os.makedirs(out_dir, exist_ok=True)
        fl_new = fl.replace(in_folder, out_folder)
        save_folder = os.path.dirname(fl_new)
        os.makedirs(save_folder, exist_ok=True)

    # Save the .zoo file
    savemat(fl_new, data)




import numpy as np


def addevent_data(data, ch, ename, etype):
    if isinstance(ch, str):
        ch = [ch]

    if len(ch) == 1 and ch[0].lower() == 'all':
        ch = [key for key in data if key != 'zoosystem']

    for channel in ch:
        if ename == '':
            data[channel]['event'] = {}
            continue

        if channel not in data:
            print(f'Channel {channel} does not exist')
            continue

        yd = data[channel]['line']  # 1D array
        etype = etype.lower()
        if etype == 'absmax':
            exd = int(np.argmax(np.abs(yd)))
            eyd = float(yd[exd])
        elif etype == 'first':
            exd = 0
            eyd = float(yd[exd])
        elif etype == 'last':
            exd = len(yd) - 1
            eyd = float(yd[exd])
        elif etype == 'max':
            exd = int(np.argmax(yd))
            eyd = float(yd[exd])
        elif etype == 'min':
            exd = int(np.argmin(yd))
            eyd = float(yd[exd])
        elif etype == 'rom':
            eyd = float(np.max(yd) - np.min(yd))
            exd = 0  # dummy index (like MATLAB version)
        else:
            raise ValueError(f'Unknown event type: {etype}')

        # Add event to the channel's event dict
        data[channel]['event'][ename] = [exd, eyd, 0]

    return data



import numpy as np
from scipy.interpolate import interp1d


def normalize_line(channel_data, nlength=101):
    """
    Channel-level: interpolate channel data to target length.
    Assumes channel_data is a 1D or 2D numpy array.
    """
    original_length = channel_data.shape[0]

    if original_length == nlength:
        return channel_data

    x_original = np.linspace(0, 1, original_length)
    x_target = np.linspace(0, 1, nlength)

    if channel_data.ndim == 1:
        f = interp1d(x_original, channel_data, kind='linear')
        channel_data_norm = f(x_target)
    else:
        channel_data_norm = np.zeros((nlength, channel_data.shape[1]))
        for i in range(channel_data.shape[1]):
            f = interp1d(x_original, channel_data[:, i], kind='linear')
            channel_data_norm[:, i] = f(x_target)

    return channel_data_norm



import numpy as np
from scipy.signal import butter, filtfilt


def filter_line(signal_raw, filt):
    """ filter an array

    Arguments
    ----------
    signal_raw : n, or n x 3 array signal to be filtered
    filt : dict, optional
        Dictionary specifying filter parameters. Keys may include:
        - 'type': 'butter' (default)
        - 'order': filter order (default: 4)
        - 'cutoff': cutoff frequency or tuple (Hz)
        - 'btype': 'low', 'high', 'bandpass', 'bandstop' (default: 'low')
        - 'fs' frequency

    Returns
    -------
    signal_filtered: filtered version of signal_raw"""
    # todo allow for missing frequency to be obtained from zoosystem metadata
    if filt is None:
        filt = {}
    if filt['type'] is 'butterworth':
        filt['type'] = 'butter'
    # Set default filter parameters
    ftype = filt.get('type', 'butter')
    order = filt.get('order', 4)
    cutoff = filt.get('cutoff', None)
    btype = filt.get('btype', 'low')
    fs = filt.get('fs', None)

    if ftype != 'butter':
        raise NotImplementedError(f"Filter type '{ftype}' not implemented.")

    if fs is None:
        raise ValueError("Sampling frequency 'fs' must be specified in filt.")

    if cutoff is None:
        raise ValueError("Cutoff frequency 'cutoff' must be specified in filt.")

    nyq = 0.5 * fs
    norm_cutoff = np.array(cutoff) / nyq

    b, a = butter(order, norm_cutoff, btype=btype, analog=False)

    if signal_raw.ndim == 1:
        signal_filtered = filtfilt(b, a, signal_raw)
    else:
        # Apply filter to each column if multivariate
        signal_filtered = np.array([filtfilt(b, a, signal_raw[:, i]) for i in range(signal_raw.shape[1])]).T

    return signal_filtered


if __name__ == '__main__':
    """ -------TESTING--------"""
    import os
    import matplotlib.pyplot as plt
    from src.biomechzoo.utils.zload import zload
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC030A05.zoo')
    data = zload(fl)
    data = data['data']
    signal_raw = data['ForceFz1']['line']
    filt = {'type': 'butterworth',
            'order': 3,
            'cutoff': 20,
            'btype': 'low',
            'fs': data['zoosystem']['Analog']['Freq']
            }
    signal_filtered = filter_line(signal_raw, filt)

    # now plot
    plt.figure(figsize=(10, 4))
    plt.plot(signal_raw, label='Raw', alpha=0.6)
    plt.plot(signal_filtered, label='Filtered', linewidth=2)
    plt.xlabel('Frame')
    plt.ylabel('Amplitude')
    plt.title('Testing filter_line')
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.show()





def split_trial(data, start_event_indx, end_event_indx):
    # todo check index problem compared to matlab start at 0 or 1
    data_new = data.copy()

    for key, value in data_new.items():
        if key == 'zoosystem':
            continue

        # Slice the line data
        data_new[key]['line'] = value['line'][start_event_indx:end_event_indx]

        # Update events if present
        if 'event' in value:
            new_events = {}
            for evt_name, evt_val in value['event'].items():
                event_frame = evt_val[0]
                # Check if event falls within the new window
                if start_event_indx <= event_frame < end_event_indx:
                    # Adjust index relative to new start
                    new_events[evt_name] = [event_frame - start_event_indx, 0, 0]
            data_new[key]['event'] = new_events

    return data_new



import numpy as np
import os
import scipy.io as sio
from biomechzoo.utils.findfield import findfield


def split_trial_by_gait_cycle(fl, event_name):
    """ splits lengthy trials containing n cycles into n trials based on side"""

    # extract folder and file name
    fld = os.path.dirname(fl)
    fl_name = os.path.splitext(os.path.basename(fl))[0]

    # load file
    data = zload(fl)

    # find all events, events should follow style name1, name2, etc..
    split_events = []
    i = 1
    _, channel_name = findfield(data, event_name)
    event_name_root = event_name[0:-1]
    while True:
        key = f"{event_name_root}{i}"
        if key in data[channel_name]['event']:
            split_events.append(data[channel_name]['event'][key])
            i += 1
        else:
            break

    n_segments = len(split_events) - 1
    if n_segments < 1:
        print("Not enough {} events to split.".format(event_name_root))
        return

    for i in range(n_segments):

        start = split_events[i][0]
        end = split_events[i + 1][0]
        fl_new = os.path.join(fld, fl_name + '_' + str(i+1) + '.zoo')
        data_new = _split_trial(fld, fl_new, start, end)

def _split_trial(fld, fl_new, start, end):
    raise NotImplementedError

if __name__ == '__main__':
    """ testing: load a single zoo file from the other subfolder in data"""
    # -------TESTING--------
    from src.biomechzoo.utils.zload import zload
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC030A05.zoo')
    split_trial_by_gait_cycle(fl, event_name='Right_FootStrike1')



import os
import pandas as pd

#todo: review. do not use in current form

def opencap2zoo_data(ik_file):
    raise NotImplementedError
    ik_data = pd.read_csv(ik_file, sep='\t', comment='#')  # OpenSim .mot format
    a = 1

def _compute_opencap_quantities():
    b = 2


if __name__ == '__main__':
    """ for unit testing"""
    # -------TESTING--------
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)

    fl = os.path.join(project_root, 'data', 'OpenCap', 'gait_3', 'OpenSimData', 'Kinematics', 'gait_3.mot')

    data = opencap2zoo_data(fl)


import numpy as np


def renamechannel_data(data, ch_old_names, ch_new_names, section='Video'):
    """
    Rename channels in a zoo data.

    Parameters
    ----------
    data : dict
        Zoo file data.
    ch_old_names : str or list
        Name of the old channels.
    ch_new_names : str or list
        Name of the new channels.

    section : str
        Section of zoo data ('Video' or 'Analog').

    Returns
    -------
    dict
        Updated zoo data with new channel added.

    Notes
    -----
    - If the channel already exists, it will be overwritten.
    - Adds channel name to the list in data['zoosystem'][section]['Channels'].
    """

    for i, ch_old_name in enumerate(ch_old_names):
        ch_new_name = ch_new_names[i]
        # Warn if overwriting
        if ch_new_name in data:
            print('Warning: channel {} already exists, overwriting...'.format(ch_new_name))

        # Assign new channel
        data[ch_new_name] = {
            'line': data[ch_old_name]['line'],
            'event': data[ch_old_name]['event']
        }

        # remove old channel
        data.pop(ch_old_name)

    # Update channel list
    ch_list = data['zoosystem'][section].get('Channels', [])

    # If the channel list is a NumPy array, convert it to a list
    if isinstance(ch_list, np.ndarray):
        ch_list = ch_list.tolist()

    # Ensure it's a flat list of strings
    if isinstance(ch_list, list) and ch_new_name not in ch_list:
        ch_list.append(ch_new_name)
        data['zoosystem'][section]['Channels'] = ch_list

    return data


if __name__ == '__main__':
    # -------TESTING--------
    import os
    from src.biomechzoo.utils.zload import zload
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC030A05.zoo')

    # load  zoo file
    data = zload(fl)
    ch_old_name = 'RKneeAngles'
    ch_new_name = 'RightKneeAngles'
    data = renamechannel_data(data, ch_old_name=ch_old_name, ch_new_name=ch_new_name)

    if ch_old_name not in data:
        print('RkneeAngles removed')
    if ch_new_name in data:
        print('RightKneeAngles added')




from biomechzoo.utils.findfield import findfield


def get_split_events(data, first_event_name):
    """ splits lengthy trials containing n cycles into n trials based on side"""

    # find all events, events should follow style name1, name2, etc..
    split_events = []
    _, channel_name = findfield(data, first_event_name)
    if channel_name is None:
        return None

    event_name_root = first_event_name[0:-1]
    first_event_number = int(first_event_name[-1])
    i = 1
    if first_event_number > 1:
        i = first_event_number

    while True:
        key = f"{event_name_root}{i}"
        if key in data[channel_name]['event']:
            split_events.append(key)
            i += 1
        else:
            break

    n_segments = len(split_events) - 1
    if n_segments < 1:
        print("Not enough {} events to split.".format(event_name_root))
        return

    return split_events




def findfield(data, target_event):
    """ searches in zoo data for the event value and channel name associated with target_event"""
    for channel, content in data.items():
        if channel == 'zoosystem':
            continue
        events = content.get('event', {})
        if target_event in events:
            return events[target_event], channel
    return None, None





import matplotlib.pyplot as plt


def zplot(data, ch, xlabel='frames', ylabel='angles (deg)'):
    """
    Plot a single channel of a zoo file, along with any existing events.

    Parameters
    ----------
    data : dict
        Loaded zoo file.
    ch : str
        Name of the channel to plot, e.g., 'RkneeAngles'.
    xlabel : str
        Label for x-axis. Default is 'frames'.
    ylabel : str
        Label for y-axis. Default is 'angles (deg)'.

    Returns
    -------
    None
    """

    if ch not in data:
        raise KeyError(f"Channel '{ch}' not found in data.")

    y = data[ch]['line']
    x = range(len(y))

    plt.figure(figsize=(10, 4))
    plt.plot(x, y, label='Signal', linewidth=2)
    plt.title(ch)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.grid(True)

    # Plot events if available
    events = data[ch].get('event', {})
    for name, coords in events.items():
        evtx, evty = coords[0], coords[1]
        plt.plot(evtx, evty, 'ro')
        plt.text(evtx, evty, name, fontsize=8, color='red', ha='left', va='bottom')

    plt.tight_layout()
    plt.show()


if __name__ == '__main__':
    # -------TESTING--------
    import os
    from biomechzoo.utils.zload import zload

    # get path to sample zoo file
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC030A05.zoo')

    # load  zoo file
    data = zload(fl)
    ch = 'SACR'
    zplot(data, ch)



import numpy as np
from biomechzoo.mvn.load_mvnx import load_mvnx


def mvnx2zoo_data(fl):
    """ loads mvnx file from xsens"""
    #todo: needs to be updated for the new version of mvnx direcly loaded in biomechzoo
    mvnx_file = load_mvnx(fl)

    # create zoo data dict
    data = {}

    # Accessing joint data :
    joint_angle_data = mvnx_file.get_joint_angle()
    joint_names = mvnx_file.joints
    for i, joint in enumerate(joint_names):
        start = i * 3
        stop = start + 3
        angles = joint_angle_data[:, start:stop]  # shape: (n_frames, 3)

        data[joint] = {
            'line': angles,
            'event': {}
        }
    # get foot strike events
    # Index 0: Left Heel contact (1 for contact, 0 for no contact)
    # Index 1: Left Toe contact (1 for contact, 0 for no contact)
    # Index 2: Right Heel contact (1 for contact, 0 for no contact)
    # Index 3: Right Toe contact (1 for contact, 0 for no contact)
    left_heel_contacts = np.array(mvnx_file.footContacts[:, 0])
    right_heel_contacts = np.array(mvnx_file.footContacts[:, 2])

    # Detect transitions from no contact (0) to contact (1)
    left_contact_start = (left_heel_contacts[:-1] == 0) & (left_heel_contacts[1:] == 1)
    right_contact_start = (right_heel_contacts[:-1] == 0) & (right_heel_contacts[1:] == 1)

    # Get indices where these transitions occur (add 1 because we're checking between frames)
    left_contact_frames = np.where(left_contact_start)[0] + 1
    right_contact_frames = np.where(right_contact_start)[0] + 1

    # add to zoo
    data['jL5S1']['event'] = {}
    for i, right_contact_frame in enumerate(right_contact_frames):
        data['jL5S1']['event']['R_FS'+str(i+1)] = [right_contact_frame, 0, 0]
    for i, left_contact_frame in enumerate(left_contact_frames):
        data['jL5S1']['event']['L_FS' + str(i + 1)] = [left_contact_frame, 0, 0]

    # add meta information
    # todo: add more, see mvnx_file object
    data['zoosystem'] = {}
    data['zoosystem']['Video'] = {}
    data['zoosystem']['Video']['Freq'] = int(mvnx_file.frame_rate)
    data['zoosystem']['Version'] = mvnx_file.version
    data['zoosystem']['configuration'] = mvnx_file.configuration
    data['zoosystem']['recording_date'] = mvnx_file.recording_date
    data['zoosystem']['original_file_name'] = mvnx_file.original_file_name
    data['zoosystem']['frame_count'] = mvnx_file.frame_count
    data['zoosystem']['comments'] = mvnx_file.comments

    return data


if __name__ == '__main__':
    """ testing """
    import os
    from src.biomechzoo.utils.zplot import zplot
    # -------TESTING--------
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    fl = os.path.join(project_root, 'data', 'other', 'Flat-001.mvnx')
    data = mvnx2zoo_data(fl)
    zplot(data, 'jRightKnee')



def explodechannel_data(data, channels=None):
    """ Explodes 3D channels (n x 3 arrays) into separate X, Y, Z channels.

    Arguments:
        data (dict): Zoo data loaded from a file
        channels (list of str or None): Channels to explode.
        If None, explode all channels with 'line' shaped (n x 3).

    Returns:
        data_new (dict): Modified zoo dictionary with exploded channels.
    """

    data_new = data.copy()

    # Find default channels if none provided
    if channels is None:
        channels = []
        for ch in data_new:
            if ch == 'zoosystem':
                continue
            ch_data = data_new[ch]['line']
            if ch_data.ndim == 2 and ch_data.shape[1] == 3:
                channels.append(ch)

    # Explode each channel
    for ch in channels:
        if ch not in data_new:
            print('Warning: channel {} not found, skipping.'.format(ch))
            continue

        ch_data = data_new[ch]['line']
        if ch_data.ndim != 2 or ch_data.shape[1] != 3:
            print(f"Warning: channel '{ch}' 'line' is not n x 3 shape, skipping.")
            continue

        x, y, z = ch_data[:, 0], ch_data[:, 1], ch_data[:, 2]
        for axis, line in zip(['_x', '_y', '_z'], [x, y, z]):
            key = ch + axis
            data_new[key] = {
                'line': line,
                'event': data[ch]['event']}

        # Remove original channel
        del data_new[ch]

    return data_new



from biomechzoo.biomech_ops.phase_angle_line import phase_angle_line
from biomechzoo.processing.addchannel_data import addchannel_data


def phase_angle_data(data, channels):
    """Compute phase angle using Hilbert Transform.
    Arguments
        data: dict, zoo data to operate on
        channels, list. Channel names on which to apply calculations
    Returns:
        data: dict, zoo data with calculations appended to new channel(s)
    """
    data_new = data.copy()
    for ch in channels:
        if ch not in data_new:
            raise ValueError('Channel {} not in data. Available keys: {}'.format(ch, list(data_new.keys())))
        r = data_new[ch]['line']
        phase_angle = phase_angle_line(r)
        ch_new = ch + '_phase_angle'
        data_new = addchannel_data(data_new, ch_new_name=ch_new, ch_new_data=phase_angle)
    return data_new


if __name__ == '__main__':
    # -------TESTING--------
    import os
    from biomechzoo.utils.zload import zload
    from biomechzoo.utils.zplot import zplot
    # get path to sample zoo file
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC032A18_exploded.zoo')

    # load  zoo file
    data = zload(fl)
    data = data['data']
    data = phase_angle_data(data, channels=['RKneeAngles_x', 'RHipAngles_x'])
    zplot(data, 'RKneeAngles_x_phase_angle')




FRAMES_ALL = -1

AXIS_ALL = -1
AXIS_X = 0
AXIS_Y = 1
AXIS_Z = 2
AXIS_W = 3  # The retrieval method will shift indices for w,x,y,z to 0,1,2,3

ANGLE_ALL = -1

ANGLE_ABDUCTION_ADDUCTION = 0
ANGLE_RADIAL_ULNAR_DEVIATION = 0
ANGLE_LATERAL_BENDING = 0

ANGLE_INTERNAL_EXTERNAL_ROTATION = 1
ANGLE_PRONATION_SUPINATION = 1
ANGLE_AXIAL_ROTATION = 1

ANGLE_FLEXION_EXTENSION = 2
ANGLE_DORSIFLEXION_PLANTARFLEXION = 2

SEGMENT_PELVIS = 0
SEGMENT_L5 = 1
SEGMENT_L3 = 2
SEGMENT_T12 = 3
SEGMENT_T8 = 4
SEGMENT_NECK = 5
SEGMENT_HEAD = 6
SEGMENT_RIGHT_SHOULDER = 7
SEGMENT_RIGHT_UPPER_ARM = 8
SEGMENT_RIGHT_FOREARM = 9
SEGMENT_RIGHT_HAND = 10
SEGMENT_LEFT_SHOULDER = 11
SEGMENT_LEFT_UPPER_ARM = 12
SEGMENT_LEFT_FOREARM = 13
SEGMENT_LEFT_HAND = 14
SEGMENT_RIGHT_UPPER_LEG = 15
SEGMENT_RIGHT_LOWER_LEG = 16
SEGMENT_RIGHT_FOOT = 17
SEGMENT_RIGHT_TOE = 18
SEGMENT_LEFT_UPPER_LEG = 19
SEGMENT_LEFT_LOWER_LEG = 20
SEGMENT_LEFT_FOOT = 21
SEGMENT_LEFT_TOE = 22

SEGMENT_CENTER_OF_MASS = 1000

SEGMENTS = {SEGMENT_PELVIS: 'Pelvis',
            SEGMENT_L5: 'L5',
            SEGMENT_L3: 'L3',
            SEGMENT_T12: 'T12',
            SEGMENT_T8: 'T8',
            SEGMENT_NECK: 'Neck',
            SEGMENT_HEAD: 'Head',
            SEGMENT_RIGHT_SHOULDER: 'RightShoulder',
            SEGMENT_RIGHT_UPPER_ARM: 'RightUpperArm',
            SEGMENT_RIGHT_FOREARM: 'RightForeArm',
            SEGMENT_RIGHT_HAND: 'RightHand',
            SEGMENT_LEFT_SHOULDER: 'LeftShoulder',
            SEGMENT_LEFT_UPPER_ARM: 'LeftUpperArm',
            SEGMENT_LEFT_FOREARM: 'LeftForeArm',
            SEGMENT_LEFT_HAND: 'LeftHand',
            SEGMENT_RIGHT_UPPER_LEG: 'RightUpperLeg',
            SEGMENT_RIGHT_LOWER_LEG: 'RightLowerLeg',
            SEGMENT_RIGHT_FOOT: 'RightFoot',
            SEGMENT_RIGHT_TOE: 'RightToe',
            SEGMENT_LEFT_UPPER_LEG: 'LeftUpperLeg',
            SEGMENT_LEFT_LOWER_LEG: 'LeftLowerLeg',
            SEGMENT_LEFT_FOOT: 'LeftFoot',
            SEGMENT_LEFT_TOE: 'LeftToe',
            SEGMENT_CENTER_OF_MASS: 'CoM'}


# regular joints

JOINT_L5_S1 = 0
JOINT_L4_L3 = 1
JOINT_L1_T12 = 2
JOINT_T9_T8 = 3
JOINT_T1_C7 = 4
JOINT_C1_HEAD = 5
JOINT_RIGHT_T4_SHOULDER = 6
JOINT_RIGHT_SHOULDER = 7
JOINT_RIGHT_ELBOW = 8
JOINT_RIGHT_WRIST = 9
JOINT_LEFT_T4_SHOULDER = 10
JOINT_LEFT_SHOULDER = 11
JOINT_LEFT_ELBOW = 12
JOINT_LEFT_WRIST = 13
JOINT_RIGHT_HIP = 14
JOINT_RIGHT_KNEE = 15
JOINT_RIGHT_ANKLE = 16
JOINT_RIGHT_BALL_FOOT = 17
JOINT_LEFT_HIP = 18
JOINT_LEFT_KNEE = 19
JOINT_LEFT_ANKLE = 20
JOINT_LEFT_BALL_FOOT = 21

JOINTS = {JOINT_L5_S1: 'jL5S1',
          JOINT_L4_L3: 'jL4L3',
          JOINT_L1_T12: 'jL1T12',
          JOINT_T9_T8: 'jT9T8',
          JOINT_T1_C7: 'jT1C7',
          JOINT_C1_HEAD: 'jC1Head',
          JOINT_RIGHT_T4_SHOULDER: 'jRightT4Shoulder',
          JOINT_RIGHT_SHOULDER: 'jRightShoulder',
          JOINT_RIGHT_ELBOW: 'jRightElbow',
          JOINT_RIGHT_WRIST: 'jRightWrist',
          JOINT_LEFT_T4_SHOULDER: 'jLeftT4Shoulder',
          JOINT_LEFT_SHOULDER: 'jLeftShoulder',
          JOINT_LEFT_ELBOW: 'jLeftElbow',
          JOINT_LEFT_WRIST: 'jLeftWrist',
          JOINT_RIGHT_HIP: 'jRightHip',
          JOINT_RIGHT_KNEE: 'jRightKnee',
          JOINT_RIGHT_ANKLE: 'jRightAnkle',
          JOINT_RIGHT_BALL_FOOT: 'jRightBallFoot',
          JOINT_LEFT_HIP: 'jLeftHip',
          JOINT_LEFT_KNEE: 'jLeftKnee',
          JOINT_LEFT_ANKLE: 'jLeftAnkle',
          JOINT_LEFT_BALL_FOOT: 'jLeftBallFoot'}

PARAMETER_JOINTS = {JOINT_L5_S1: 'j_l5_s1',
                    JOINT_L4_L3: 'j_l4_l3',
                    JOINT_L1_T12: 'j_l1_t12',
                    JOINT_T9_T8: 'j_t9_t8',
                    JOINT_T1_C7: 'j_t1_c7',
                    JOINT_C1_HEAD: 'j_c1_head',
                    JOINT_RIGHT_T4_SHOULDER: 'j_right_t4_shoulder',
                    JOINT_RIGHT_SHOULDER: 'j_right_shoulder',
                    JOINT_RIGHT_ELBOW: 'j_right_elbow',
                    JOINT_RIGHT_WRIST: 'j_right_wrist',
                    JOINT_LEFT_T4_SHOULDER: 'j_left_t4_shoulder',
                    JOINT_LEFT_SHOULDER: 'j_left_shoulder',
                    JOINT_LEFT_ELBOW: 'j_left_elbow',
                    JOINT_LEFT_WRIST: 'j_left_wrist',
                    JOINT_RIGHT_HIP: 'j_right_hip',
                    JOINT_RIGHT_KNEE: 'j_right_knee',
                    JOINT_RIGHT_ANKLE: 'j_right_ankle',
                    JOINT_RIGHT_BALL_FOOT: 'j_right_ball_foot',
                    JOINT_LEFT_HIP: 'j_left_hip',
                    JOINT_LEFT_KNEE: 'j_left_knee',
                    JOINT_LEFT_ANKLE: 'j_left_ankle',
                    JOINT_LEFT_BALL_FOOT: 'j_left_ball_foot'}


# ergonomic joints

ERGO_JOINT_T8_HEAD = 0
ERGO_JOINT_T8_LEFT_UPPER_ARM = 1
ERGO_JOINT_T8_RIGHT_UPPER_ARM = 2
ERGO_JOINT_PELVIS_T8 = 3
ERGO_JOINT_VERTICAL_PELVIS = 4
ERGO_JOINT_VERTICAL_T8 = 5

ERGO_JOINT_INDICES = {ERGO_JOINT_T8_HEAD,
                      ERGO_JOINT_T8_LEFT_UPPER_ARM,
                      ERGO_JOINT_T8_RIGHT_UPPER_ARM,
                      ERGO_JOINT_PELVIS_T8,
                      ERGO_JOINT_VERTICAL_PELVIS,
                      ERGO_JOINT_VERTICAL_T8}

ERGO_JOINTS = {ERGO_JOINT_T8_HEAD: 'T8_Head',
               ERGO_JOINT_T8_LEFT_UPPER_ARM: 'T8_LeftUpperArm',
               ERGO_JOINT_T8_RIGHT_UPPER_ARM: 'T8_RightUpperArm',
               ERGO_JOINT_PELVIS_T8: 'Pelvis_T8',
               ERGO_JOINT_VERTICAL_PELVIS: 'Vertical_Pelvis',
               ERGO_JOINT_VERTICAL_T8: 'Vertical_T8'}

ERGO_PARAMETER_JOINTS = {ERGO_JOINT_T8_HEAD: 't8_head',
                         ERGO_JOINT_T8_LEFT_UPPER_ARM: 't8_left_upper_arm',
                         ERGO_JOINT_T8_RIGHT_UPPER_ARM: 't8_right_upper_arm',
                         ERGO_JOINT_PELVIS_T8: 'pelvis_t8',
                         ERGO_JOINT_VERTICAL_PELVIS: 'vertical_p1elvis',
                         ERGO_JOINT_VERTICAL_T8: 'vertical_t8'}


# foot contacts

FOOT_CONTACT_LEFT_HEEL = 1
FOOT_CONTACT_LEFT_TOE = 2
FOOT_CONTACT_RIGHT_HEEL = 4
FOOT_CONTACT_RIGHT_TOE = 8

# some points

POINT_JOINT_ANKLE = 0
POINT_JOINT_BALL_FOOT = 1
POINT_HEEL_FOOT = 2
POINT_FIRST_METATARSAL = 3
POINT_FIFTH_METATARSAL = 4
POINT_PIVOT_FOOT = 5
POINT_HEEL_CENTER = 6
POINT_TOP_OF_FOOT = 7

POINTS_LEFT_FOOT = {
    POINT_JOINT_ANKLE: 'jLeftAnkle',
    POINT_JOINT_BALL_FOOT: 'jLeftBallFoot',
    POINT_HEEL_FOOT: 'pLeftHeelFoot',
    POINT_FIRST_METATARSAL: 'pLeftFirstMetatarsal',
    POINT_FIFTH_METATARSAL: 'pLeftFifthMetatarsal',
    POINT_PIVOT_FOOT: 'pLeftPivotFoot',
    POINT_HEEL_CENTER: 'pLeftHeelCenter',
    POINT_TOP_OF_FOOT: 'pLeftTopOfFoot',
}

POINTS_RIGHT_FOOT = {
    POINT_JOINT_ANKLE: 'jRightAnkle',
    POINT_JOINT_BALL_FOOT: 'jRightBallFoot',
    POINT_HEEL_FOOT: 'pRightHeelFoot',
    POINT_FIRST_METATARSAL: 'pRightFirstMetatarsal',
    POINT_FIFTH_METATARSAL: 'pRightFifthMetatarsal',
    POINT_PIVOT_FOOT: 'pRightPivotFoot',
    POINT_HEEL_CENTER: 'pRightHeelCenter',
    POINT_TOP_OF_FOOT: 'pRightTopOfFoot',
}

# Toe
# POINT_JOINT_BALL_FOOT = 0 # overlaps with foot segment
POINT_TOE = 1
POINT_JOINT_TOE = 2

POINTS_LEFT_TOE = {
    # POINT_JOINT_BALL_FOOT: 'jLeftBallFoot',
    POINT_TOE: 'pLeftToe',
    POINT_JOINT_TOE: 'jLeftToe',
}

POINTS_RIGHT_TOE = {
    # POINT_JOINT_BALL_FOOT: 'jRightBallFoot',
    POINT_TOE: 'pRightToe',
    POINT_JOINT_TOE: 'jRightToe',
}



from biomechzoo.utils.findfield import findfield
import warnings


def partition_data(data, evt_start, evt_end):
    """ partition data for all channels between events evt_start and evt_end"""

    # extract event values
    e1, _ = findfield(data,evt_start)
    e2, _ = findfield(data,evt_end)

    data_new = data.copy()
    for ch_name, ch_data in data_new.items():
        if ch_name != 'zoosystem':
            try:
                data_new[ch_name]['line'] = ch_data[e1[0]:e2[0], :]
            except (IndexError, ValueError) as e:
                # IndexError: if e1[0]:e2[0] goes beyond the available indices
                # ValueError: less likely, but may arise with shape mismatches
                warnings.warn(f"Skipping {ch_name} due to error: {e}")

            # partition events
            events = ch_data['event']
            for event_name, value in events.items():
                original_frame = value[0]
                if original_frame == 1:
                    continue  # leave index 1 as is (per MATLAB version)
                elif original_frame == 999:
                    continue  # do not change outlier markers
                else:
                    new_frame = original_frame - e1[0] + 1
                    data_new[ch_name]['event'][event_name][0] = new_frame

    return data_new


def _partition_line(arr, evt_start, evt_end):
    arr_new = arr[evt_start:evt_end, :]
    return arr_new


def _partition_event(event_dict, evt_start, evt_end, arr_len):
    raise NotImplementedError
    # event_dict_new = {}
    # for event, event_val in event_dict:
    #     event_val_new =
    #     event_dict_new[event] =
    #
    #
    #
    # return event_dict_new



import warnings
from biomechzoo.biomech_ops.normalize_line import normalize_line


def normalize_data(data, nlength=101):
    """normalize all channels in the loaded zoo dict to nlen.
    Arguments
        data: dict, loaded zoo file
        nlength: int: new length of data. Default = 101, usually a movement cycle
    Returns:
        None
    Notes:
        -It is often needed to partition data to a single cycle first (see partition_data)
    """

    # normalize channel length
    data_new = data.copy()
    for ch_name, ch_data in data_new.items():
        if ch_name != 'zoosystem':
            ch_data_line = ch_data['line']
            ch_data_event = ch_data['event']
            ch_data_normalized = normalize_line(ch_data_line, nlength)
    data_new[ch_name]['line'] = ch_data_normalized
    data_new[ch_name]['event'] = ch_data_event
    warnings.warn('event data have not been normalized')

    # update zoosystem
    # todo: update all relevant zoosystem meta data related to data lengths
    warnings.warn('zoosystem data have not been fully updated')
    if 'Video' in data['zoosystem']:
        data['zoosystem']['Video']['CURRENT_END_FRAME'] = nlength
    if 'Analog' in data['zoosystem']:
        data['zoosystem']['Analog']['CURRENT_END_FRAME'] = nlength

    return data_new



import pandas as pd
import os
import re

from biomechzoo.utils.compute_sampling_rate_from_time import compute_sampling_rate_from_time


def csv2zoo_data(csv_path, header_len=10):

    # Read header lines until 'endheader'
    header_lines = []
    with open(csv_path, 'r') as f:
        for line in f:
            header_lines.append(line.strip())
            if line.strip().lower() == 'endheader':
                break

    # Parse metadata
    metadata = _parse_metadata(header_lines)

    # Step 3: Load data
    df = pd.read_csv(csv_path, skiprows=header_len)
    time = df.iloc[:, 0].values  # first column is Time
    data = df.iloc[:, 1:]

    # S Assemble zoo data
    zoo_data = {}
    for ch in data.columns:
        zoo_data[ch] = {
            'line': data[ch].values,
            'event': []
        }

    # compute sampling rate
    fsamp = compute_sampling_rate_from_time(time)

    # add metadata
    # todo update zoosystem to match biomechzoo requirements
    zoo_data['zoosystem'] = metadata
    zoo_data['zoosystem']['Freq'] = fsamp

    return zoo_data


def _parse_metadata(header_lines):
    metadata = {}
    for line in header_lines:
        if '=' in line:
            key, val = line.split('=', 1)
            key = key.strip()
            val = val.strip()

            # Strip trailing commas and whitespace explicitly
            val = val.rstrip(',').strip()

            # Extract first numeric token if any
            match = re.search(r'[-+]?\d*\.?\d+', val)
            if match:
                num_str = match.group(0)
                try:
                    val_num = int(num_str)
                except ValueError:
                    val_num = float(num_str)
            else:
                # Now val should be clean of trailing commas, so just lower case it
                val_num = val.lower()

            metadata[key] = val_num
    return metadata


if __name__ == '__main__':
    """ for unit testing"""
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    csv_file = os.path.join(project_root, 'data', 'other', 'opencap_walking1.csv')

    data = csv2zoo_data(csv_file)
